module Isomorfeus
  module Speednode
    class Runtime < ExecJS::Runtime
      def self.attach_proc(context_id, func, run_block)
        attached_procs[context_id] = { func => run_block }
      end

      def self.attached_procs
        @attached_procs ||= {}
      end

      def self.responders
        @responders ||= {}
      end

      attr_reader :name, :vm

      def initialize(options)
        @name        = options[:name]
        @binary      = Isomorfeus::Speednode::NodeCommand.cached(options[:command])
        @runner_path = options[:runner_path]
        @encoding    = options[:encoding]
        @deprecated  = !!options[:deprecated]

        @vm = VM.new(
          binary: @binary,
          source_maps: '--enable-source-maps',
          runner_path: @runner_path
        )

        @popen_options = {}
        @popen_options[:external_encoding] = @encoding if @encoding
        @popen_options[:internal_encoding] = ::Encoding.default_internal || 'UTF-8'
      end

      def available?
        @binary ? true : false
      end

      def deprecated?
        @deprecated
      end
    end
  end
end
